/*************************************************************************
** interpcom-2.4    (command interpreter)                                **
** intl_.c : Messages and command names                                  **
**                                                                       **
** Copyright (C) 2001  Jean-Marc Drezet                                  **
**                                                                       **
**  This library is free software; you can redistribute it and/or        **
**  modify it under the terms of the GNU Library General Public          **
**  License as published by the Free Software Foundation; either         **
**  version 2 of the License, or (at your option) any later version.     **
**                                                                       **
**  This library is distributed in the hope that it will be useful,      **
**  but WITHOUT ANY WARRANTY; without even the implied warranty of       **
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    **
**  Library General Public License for more details.                     **
**                                                                       **
**  You should have received a copy of the GNU Library General Public    **
**  License along with this library; if not, write to the Free           **
**  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.   **
**                                                                       **
** Please mail any bug reports/fixes/enhancements to me at:              **
**      drezet@math.jussieu.fr                                           **
** or                                                                    **
**      Jean-Marc Drezet                                                 **
**      Institut de Mathematiques                                        **
**      UMR 7586 du CNRS                                                 **
**      173, rue du Chevaleret                                           **
**      75013 Paris                                                      **
**      France                                                           **
**                                                                       **
 *************************************************************************/

/*--------------------------------------------------------------------
        English
--------------------------------------------------------------------*/
#ifdef _ENG_LANG
char           *_INTCOM_colors[] = {
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "violet",
    "lightblue",
    "white",
};

char           *mess_interp[] =
{
    "Missing parameter",                                      /* 0    */
    "Incorrect file name",                                    /* 1    */
    "Incorrect numerical parameter",                          /* 2    */
    "Incorrect second parameter",                             /* 3    */
    "Too many conditional variables",                         /* 4    */
    "Too many open files",                                    /* 5    */
    "Non opened file",                                        /* 6    */
    "Incorrect question file",                                /* 7    */
    "Incorrect object name",                                  /* 8    */
    "Too long object name",                                   /* 9    */
    "Too many objects of this type",                          /* 10   */
    "Already used name",                                      /* 11   */
    "Unknown object",                                         /* 12   */
    "Incorrect number of object or structure types",          /* 13   */
    "Incorrect objects",                                      /* 14   */
    "Incorrect saved object",                                 /* 15   */
    "Too long structure name",                                /* 16   */
    "Too many structures of this type",                       /* 17   */
    "Unknown structure",                                      /* 18   */
    "Incorrect member name",                                  /* 19   */
    "Incorrect command file name",                            /* 20   */
    "Incorrect structure name",                               /* 21   */
    "Incorrect program file",                                 /* 22   */
    "Incorrect program file name",                            /* 23   */
    "Unknown program",                                        /* 24   */
    "Incorrect command in this context",                      /* 25   */
    "Already used program name",                              /* 26   */
    "Too many lines",                                         /* 27   */
    "Unknown program",                                        /* 28   */
    "File already open for reading",                          /* 29   */
    "File already open for writing",                          /* 30   */
    "Too many threads",                                       /* 31   */
    "Already used thread name",                               /* 32   */
    "Incorrect first parameter",                              /* 33   */
    "Too long directory name",                                /* 34   */
    "Non global variable",                                    /* 35   */
    "Only the main thread is running",                        /* 36   */
    "Too many string variables",                              /* 37   */
    "Unknown string variable",                                /* 38   */
    "This command can be used only in the main thread",       /* 39   */
    "Other threads are running",                              /* 40   */
    "Already used string variable",                           /* 41   */
    "Initialization file not found",                          /* 42   */
    "Incorrect program name",                                 /* 43   */
    "Incorrect program",                                      /* 44   */
    "Incorrect number of parameters",                         /* 45   */
    "Already used command name",                              /* 46   */
    "Incorrect parameter",                                    /* 47   */
    "Incorrect type of variable",                             /* 48   */
    "Too many object types",                                  /* 49   */
    "Incorrect number of dimensions",                         /* 50   */
    "Incorrect maximal number of objects",                    /* 51   */
    "Alias of an undefined object type",                      /* 52   */
    "Too many structure types",                               /* 53   */
    "Incorrect number of members",                            /* 54   */
    "Unknown variable",                                       /* 55   */
    "Unknown thread",                                         /* 56   */
    "Type 'exit' to kill the main thread",                    /* 57   */
    "Objects of undefined data type",                         /* 58   */
    "Incorrect object type(s)",                               /* 59   */
    "Incorrect expression evaluator",                         /* 60   */
    "Unknown color",                                          /* 61   */

    ""
};

/* English command names */
char           *names_interp[]  =
{
    "add",
    "assign",
    "ch_expr",
    "clock",
    "close",
    "cmultiply",
    "color_echo",
    "conjg",
    "const",
    "const_c",
    "const_r",
    "const_th",
    "copy",
    "delcom",
    "deldat",
    "delprog",
    "delres",
    "delstring",
    "desassign",
    "desc",
    "destroy",
    "echo",
    "echof",
    "echoi",
    "editor",
    "end_mon",
    "eval",
    "exit",
    "fillobj",
    "flush",
    "greet",
    "hist",
    "Imag",
    "initvar",
    "is",
#ifdef _HAVE_THREADS
    "kill",
#endif
    "list",
    "liststr",
#ifdef _HAVE_THREADS
    "listt",
#endif
    "load",
    "loop",
    "loopf",
    "mkcomplex",
    "mkpolaire",
    "mon",
    "multiply",
    "numcom",
    "objdef",
#ifdef _HAVE_THREADS
    "pause",
#endif
    "phase",
    "proglist",
    "question",
    "radius",
    "read",
    "Real",
    "repete",
    "restore",
    "send_print",
    "setdir",
    "setobj",
    "shell",
    "si",
    "silence",
    "string",
    "strucdef",
    "substract",
    "svg",
#ifdef _HAVE_THREADS
    "thread",
#endif
    "time",
    "undef",
    "varlist",
#ifdef _HAVE_THREADS
    "wait",
#endif
    "write",
};


/* Expression evaluator error messages */
char           *_EXPREV_errmessg[] = {
    "Syntax error",
    "Unbalanced parenthesis",
    "Division by zero",
    "Unknown variable",
    "Too many variables",
    "Unknown function",
    "Wrong number of arguments",
    "Wrong number of arguments",
};

char           *_TIME_messg[] = {
    " ; Execution time : ",
    "Time : ",
};

char           *_PROGREPL_messg[] = {
    "Program ",
    "replaced ",
};

char           *_INITERR_messg[] = {
    "Error in the section ",
    " of the initialization file",
    "Error in the initialization file",
    "Line ",
};

char           *_DELPROG_messg[] = {
    "Destruction of program ",
    " not allowed",
};

char *_MEMMC_messg[] = {
    "Incorrect arrays",
    "Incorrect stored array",
};
#endif
/*------------------------------------------------------------------*/




/*--------------------------------------------------------------------
    French
--------------------------------------------------------------------*/
#ifdef _FR_LANG
char           *_INTCOM_colors[] = {
    "noir",
    "rouge",
    "vert",
    "jaune",
    "bleu",
    "violet",
    "bleuclair",
    "blanc",
};

char           *mess_interp[] =
{
    "Parametre manquant",                                     /* 0    */
    "Nom de fichier incorrect",                               /* 1    */
    "Parametre numerique incorrect",                          /* 2    */
    "Second parametre incorrect",                             /* 3    */
    "Trop de variables conditionnelles",                      /* 4    */
    "Trop de fichiers ouverts",                               /* 5    */
    "Fichier non ouvert",                                     /* 6    */
    "Fichier de questions incorrect",                         /* 7    */
    "Nom d'objet incorrect",                                  /* 8    */
    "Nom d'objet trop long",                                  /* 9    */
    "Trop d'objets de ce type definis",                       /* 10   */
    "Nom deja utilise",                                       /* 11   */
    "Objet inexistant",                                       /* 12   */
    "Numero de type d'objet ou de structure incorrect",       /* 13   */
    "Objets incorrects",                                      /* 14   */
    "Objet enregistre incorrect",                             /* 15   */
    "Nom de structure trop long",                             /* 16   */
    "Trop de structures de ce type definies",                 /* 17   */
    "Structure inexistante",                                  /* 18   */
    "Nom de membre incorrect",                                /* 19   */
    "Nom de fichier de commandes incorrect\n",                /* 20   */
    "Nom de structure incorrect",                             /* 21   */
    "Fichier de programme incorrect",                         /* 22   */
    "Nom de fichier de programme incorrect",                  /* 23   */
    "Programme inconnu",                                      /* 24   */
    "Commande incorrecte dans ce contexte",                   /* 25   */
    "Nom de programme deja utilise",                          /* 26   */
    "Nombre de lignes trop grand",                            /* 27   */
    "Programme inexistant",                                   /* 28   */
    "Fichier deja ouvert en lecture",                         /* 29   */
    "Fichier deja ouvert en ecriture",                        /* 30   */
    "Trop de threads",                                        /* 31   */
    "Nom de thread deja utilise",                             /* 32   */
    "Premier parametre incorrect",                            /* 33   */
    "Nom de repertoire trop long",                            /* 34   */
    "Variable non globale",                                   /* 35   */
    "Seul le thread principal existe",                        /* 36   */
    "Trop de chaines de caracteres",                          /* 37   */
    "Chaine de caracteres inconnues",                         /* 38   */
    "Commande utilisable seulement dans le thread principal", /* 39   */
    "D'autres threads sont en cours d'execution",             /* 40   */
    "Nom de chaine de caracteres deja utilise",               /* 41   */
    "Fichier d'initialisation non trouve",                    /* 42   */
    "Nom de programme incorrect",                             /* 43   */
    "Programme incorrect",                                    /* 44   */
    "Nombre de parametres incorrect",                         /* 45   */
    "Nom de commande deja utilise",                           /* 46   */
    "Parametre incorrect",                                    /* 47   */
    "Type de variable incorrect",                             /* 48   */
    "Trop de types d'objet"                                   /* 49   */
    "Nombre de dimensions incorrect",                         /* 50   */
    "Nombre maximal d'objets incorrect",                      /* 51   */
    "Alias d'un type d'objet inconnu",                        /* 52   */
    "Trop de types de structure",                             /* 53   */
    "Nombre de membres incorrect",                            /* 54   */
    "Variable inconnue",                                      /* 55   */
    "Thread inconnu",                                         /* 56   */
    "Faire 'exit' pour tuer le thread main",                  /* 57   */
    "Objets de type indetermine",                             /* 58   */
    "Type d'objet incorrect",                                 /* 59   */
    "Evaluateur d'expressions incorrect",                     /* 60   */
    "Couleur inconnue",                                       /* 61   */

    ""
};

/* French command names */
char           *names_interp[]  =
{
    "ajoute_",
    "assign",
    "ch_expr",
    "horloge",
    "close",
    "cmultiplie",
    "color_echo",
    "conjg",
    "const",
    "const_c",
    "const_r",
    "const_th",
    "copie",
    "delcom",
    "deldat",
    "delprog",
    "delres",
    "delstring",
    "desassign",
    "desc",
    "detruit",
    "echo",
    "echof",
    "echoi",
    "editeur",
    "fin_mon",
    "eval",
    "exit",
    "fillobj",
    "flush",
    "greet",
    "hist",
    "Imag",
    "initvar",
    "is",
#ifdef _HAVE_THREADS
    "kill",
#endif
    "list",
    "liststr",
#ifdef _HAVE_THREADS
    "listt",
#endif
    "load",
    "boucle",
    "bouclef",
    "mkcomplex",
    "mkpolaire",
    "mon",
    "multiplie_",
    "numcom",
    "objdef",
#ifdef _HAVE_THREADS
    "pause",
#endif
    "phase",
    "progliste",
    "question",
    "radius",
    "lit",
    "Real",
    "repete",
    "restore",
    "send_print",
    "setdir",
    "setobj",
    "shell",
    "si",
    "silence",
    "string",
    "strucdef",
    "soustrait_",
    "svg",
#ifdef _HAVE_THREADS
    "thread",
#endif
    "time",
    "undef",
    "varlist",
#ifdef _HAVE_THREADS
    "attends",
#endif
    "ecrit",
};

/* Expression evaluator error messages */
char           *_EXPREV_errmessg[] = {
    "Erreur de syntaxe",
    "Probleme de parenthese",
    "Division par zero",
    "Variable inconnue",
    "Trop de variables",
    "Fonction inconnue",
    "Nombre d'arguments incorrect",
    "Argument manquant",
};

char           *_TIME_messg[] = {
    " ; Temps d'execution : ",
    "Temps : ",
};

char           *_PROGREPL_messg[] = {
    "Programme ",
    "remplace ",
};

char           *_INITERR_messg[] = {
    "Erreur dans la section ",
    " du fichier d'initialisation",
    "Erreur dans le fichier d'initialisation",
    "Ligne ",
};

char           *_DELPROG_messg[] = {
    "Destruction du programme ",
    " interdite",
};

char *_MEMMC_messg[] = {
    "Tableaux incorrects",
    "Tableau enregistre incorrect",
};
#endif
/*------------------------------------------------------------------*/
